//=============================================================================
// File Name: EGIPsolver.cc
// Written by: Shai Shalev-Shwartz 
//
// Implement a solver for the optimiztion problem:
// Given vectors P,Q , constant gamma, and index y, find:
//   min_{x}  \sum_r log(exp(x_r)*P(r) + Q(r)) -  gamma * x_{y}
//    s.t.
//        x_{y} \in [0,C]
//        sum_r x_r = 0
//        forall r \neq y, x_r <= 0
//=============================================================================

//=============================================================================
// Included files
//=============================================================================
#include <iostream>
#include <fstream>
#include <stdio.h>
#include <cstdlib>
#include <cmdline/cmd_line.h>
#include <kernels/kernels.h>
#include <kernels/kernels.imp>
#include <infra.h>


// accuracy of the IP algorithm
#define IP_EPSI 1e-8


// declarations of some help functions
void doubleswap(double &a,double &b);
double calculate_rt(infra::vector_base PQ, infra::vector_base x, infra::vector_base lambda,
		    double lamp, double lamn, double nu, 
		    double t, double gamma, double C);
bool satisfy_inequality_constraints(infra::vector_base x,double C);



/** This function implements a solver for the optimiztion problem:
    Given vectors P,Q , constant gamma, and index y, find:
    min_{x}  \sum_r log(exp(x_r)*P(r) + Q(r)) -  gamma * x_{y}
     s.t.
      x_{y} \in [0,C]
      sum_r x_r = 0
      forall r \neq y, x_r <= 0
    @param P a vector_base of the coefficients P(r)
    @param Q a vector_base of the coefficients Q(r)
    @param gamma a double (in the definitiobn of the objective)
    @param C a double (in the definition of the constraints)
    @param y an index of the unique index
    @param x this vector contains initial solution and will
             contain the optimum after the function returns
*/
void EG_IPsolver(infra::vector_base P, infra::vector_base Q, 
		 double gamma, double C, unsigned int y,
		 infra::vector_base x) 
{
  
  // initialize
  // ==========

  uint n = x.size(); // number of variables
  uint m = n+1; // number of inequality constraints

  // for conveniency, we put the index y at the end of the vector
  if (y != n-1) {
    uint last_ind = n-1;
    doubleswap(x(y),x(last_ind));
    doubleswap(P(y),P(last_ind));
    doubleswap(Q(y),Q(last_ind));
  }

  // it helps to have references to the prefix of x and to the last element of x
  infra::vector_base prefix_x = x.subvector(0,n-1);
  double& last_x = x(n-1);

  // for efficiency, calculate the ratio between Q and P
  infra::vector PQ(Q); PQ /= P;
  infra::vector_base prefix_PQ = PQ.subvector(0,n-1);
  double& last_PQ = PQ(n-1);

    
  // move x from the boundaries
  if (last_x > C-1e-6) last_x = C-1e-6;
  if (last_x < 1e-6) last_x = 1e-6;
  for (uint i=0; i < n-1; ++i) {
    if (x(i) > -1e-6) x(i) = -1e-6;
  }

  // make sure that x.sum() is zero
  double tmp = -prefix_x.sum();
  if (tmp != last_x) {
    prefix_x *= (last_x/tmp);
  }

  // set the lagrange multipliers
  infra::vector lambda(n-1); // for the constraints prefix_x <= 0
  double lamp; // for last_x >= 0, or equivalently, -last_x <= 0
  double lamn = 0.0; // for last_x - C <= 0
  double nu = 0.0; // for x.sum() = 0

  // initialize lagrange multipliers so that r_dual will be zero
  // that is:  
  infra::vector delta_f0(n-1);
  for (uint i=0; i<n-1; ++i) {
    delta_f0(i) = 1 / (1 + PQ(i)*exp(-x(i)));
  }
  nu  = - delta_f0.max();
  lambda = -nu; lambda -= delta_f0;
  double last_delta_f0_plus_nu = 1 / (1 + last_PQ*exp(-last_x)) - gamma + nu;
  if (last_delta_f0_plus_nu > 0.0) {
    lamp = last_delta_f0_plus_nu;
    lamn = 0.0; 
  } else {
    lamp = 0.0;
    lamn = - last_delta_f0_plus_nu;
  }

  lambda += 1e-3; // make sure that lambda is not zero
  lamp += 1e-3; lamn += 1e-3;

  // find the surrogate duality gap
  double eta_hat = - (prefix_x*lambda) + last_x*lamp - (last_x-C)*lamn;

  // IP method parameters (these are constants)
  double epsi = IP_EPSI;
  double epsi_feas = epsi;
  double MU = 10.0;
  double BETA = 0.5;
  double ALPHA = 0.01;

  // Define some objects that will be used in the algorithm
  infra::vector invH(n); 
  infra::vector v(n);
  infra::vector delta_x(n);
  infra::vector delta_lambda(n-1);


  // ===============================================================
  // IP iterations
  // ===============================================================
  uint MAX_ITER = 100;  // perform at most MAX_ITER iterstions
  uint iter = 0;
  for ( ; iter < MAX_ITER; ++iter) {

    // ===================
    // step 1: determine t
    // ===================

    double t = MU*m/eta_hat;
    //std::cerr << "eta_hat = " << eta_hat << " t = " << t << std::endl;
    
    // ============================================
    // step 2: Compute primal-dual search direction
    // ============================================
    
    // calculate the inverse of the diagonal of H_pd
    for (uint i=0; i<n-1 ; ++i) 
      invH(i) = (PQ(i)*exp(-x(i))) / ( (1+PQ(i)*exp(-x(i))) * (1+PQ(i)*exp(-x(i))) ) 	- (lambda(i)/x(i));
    invH(n-1) = (last_PQ*exp(-last_x)) / ( (1+last_PQ*exp(-last_x)) * (1+last_PQ*exp(-last_x)) )
      - (lamp/(-last_x)) - (lamn/(last_x - C));
    for (uint i=0; i<n; ++i) invH(i) = 1/invH(i);
    

    // calculate the target vector for the Newton step
    for (uint i=0; i<n-1 ; ++i) 
      v(i) = - ( 1 / (1+PQ(i)*exp(-x(i))) - 1 / (t*x(i)) + nu  );
    v(n-1) = - ( 1 / (1+last_PQ*exp(-last_x)) - gamma - 1 / (t*last_x) - 1 / (t*(last_x-C)) + nu  );
    
    // calculate r_pri
    double r_pri = x.sum();
    double minus_r_pri = -r_pri;

    // now we can easily find delta_nu
    double delta_nu = -(minus_r_pri - invH*v) / invH.sum();

    // and from delta_nu we can find delta_x = invH*(v - delta_nu)
    delta_x = v; delta_x -= delta_nu; delta_x *= invH;

    // finally, we find delta_lambda, delta_lamp, and delta_lamn
    for (uint i=0; i<n-1; ++i) {
      delta_lambda(i) = - lambda(i) - 1/(t*x(i)) - delta_x(i)*lambda(i)/x(i);
    }
    double delta_lamp = - lamp  - 1/(t*(-last_x)) - delta_x(n-1)*lamp*(-1.0)/(-last_x);
    double delta_lamn = - lamn - 1/(t*(last_x-C)) - delta_x(n-1)*lamn/(last_x - C);


#ifdef NO_DEF
    // sanity check
    double max_diff = 0.0;
    for (uint i=0; i<n-1; ++i) {
      double left_hand = PQ(i)*exp(-x(i)) / ( (1+PQ(i)*exp(-x(i))) * (1+PQ(i)*exp(-x(i))) ) * delta_x(i)
	+ delta_lambda(i) + delta_nu;
      double right_hand = - (1 / (1+PQ(i)*exp(-x(i))) + lambda(i) + nu  ); 
      if (fabs(left_hand-right_hand) > max_diff) max_diff = fabs(left_hand-right_hand);
    }
    double left_hand = last_PQ*exp(-last_x) / ( (1+last_PQ*exp(-last_x)) * (1+last_PQ*exp(-last_x)) ) * delta_x(n-1)
      - delta_lamp + delta_lamn + delta_nu;
    double right_hand = - (1 / (1+last_PQ*exp(-last_x)) - gamma - lamp + lamn + nu  );
    if (fabs(left_hand-right_hand) > max_diff) max_diff = fabs(left_hand-right_hand);
    std::cerr << "max-diff = " << max_diff << std::endl;
    std::cerr << "lala = " << 1/invH(n-1) * delta_x(n-1) + delta_nu - v(n-1) << std::endl;;
#endif

					      
    // ==============================
    // step 3: Line serach and update
    // ==============================

    // first find s = 0.99 * sup { s \in [0,1] : lamp+s*delta_p > 0 and lamn+s*delta_n > 0}
    double s = 1.0;
    for (uint i=0; i<n-1; ++i) {
      double tmp = (-lambda(i)/delta_lambda(i));
      if (delta_lambda(i) < 0 &&  tmp < s) s = tmp;
    }
    double tmp = (-lamp/delta_lamp);
    if (delta_lamp < 0 &&  tmp < s) s = tmp;
    tmp = (-lamn/delta_lamn);
    if (delta_lamn < 0 &&  tmp < s) s = tmp;

    s *= 0.99;

    // keep multiplying s *= BETA until x+s*delta_x satisfies 
    // all the inequality constraints

    while (!satisfy_inequality_constraints(x+s*delta_x,C))
      s *= BETA;


    // Now, continue multiplying s by BETA until ||rtp|| <= (1-ALPHA*s) ||rt|| where,
    // ||rt||^2 = ||r_dual||^2 + ||r_cent||^2 + ||r_pri||^2
    // ||rtp|| is like ||rt|| just that all variables are incremented by s*delta_?
    double rt = calculate_rt(PQ,x,lambda,lamp,lamn,nu,t,gamma,C);
    double rtp = calculate_rt(PQ,x+s*delta_x,lambda+s*delta_lambda,lamp+s*delta_lamp,
			      lamn+s*delta_lamn,nu+s*delta_nu,t,gamma,C);
    while (rtp > (1-ALPHA*s)*rt) {
        s *= BETA;
	rtp = calculate_rt(PQ,x+s*delta_x,lambda+s*delta_lambda,lamp+s*delta_lamp,
			   lamn+s*delta_lamn,nu+s*delta_nu,t,gamma,C);
    }


    // now update
    x += s*delta_x;
    lambda += s*delta_lambda; lamp += s*delta_lamp; lamn += s*delta_lamn;
    nu += s*delta_nu;
    
    // stopping criterion
    eta_hat = - (prefix_x*lambda) + last_x*lamp - (last_x-C)*lamn;
    if (eta_hat <= epsi) { // if surrogate gap is small
      double tmp = x.sum();
      if (fabs(tmp) <= epsi_feas) { // if primal residual is small
	double rt = 0.0;
	for (uint i=0; i<n-1; ++i) {
	  tmp = 1/(1+PQ(i)*exp(-x(i))) + lambda(i) + nu;
	  rt += tmp*tmp;
	}
	tmp = 1/(1+PQ(n-1)*exp(-x(n-1))) - gamma - lamp + lamn + nu;
	rt += tmp*tmp;

	if (sqrt(rt) <= epsi_feas) { // if dual residual is small
	  //std::cerr << "Converged after " << iter << " iterations" << std::endl;
	  break;
	}
      }
    }
  }

  if (iter == MAX_ITER) {
    std::cerr << "Warning: IPsolver had not converged." << std::endl;
  }


  // put back the index y at its place 
  if (y != n-1) {
    uint last_ind = n-1;
    doubleswap(x(y),x(last_ind));
  }

  // that's it !


}



// this function swaps a and b
void doubleswap(double &a,double &b) {
  double tmp = a;
  a = b;
  b = tmp;
}


// this function returns ||rt|| where,
// ||rt||^2 = ||r_dual||^2 + ||r_pri||^2 + ||r_cent||^2
double calculate_rt(infra::vector_base PQ, infra::vector_base x, infra::vector_base lambda,
		    double lamp, double lamn, double nu, 
		    double t, double gamma, double C) {

  uint n = x.size();
  double rt = 0.0;
  double tmp;

  // add ||r_dual||^2
  for (uint i=0; i<n-1; ++i) {
    tmp = 1/(1+PQ(i)*exp(-x(i))) + lambda(i) + nu;
    rt += tmp*tmp;
  }
  tmp = 1/(1+PQ(n-1)*exp(-x(n-1))) - gamma - lamp + lamn + nu;
  rt += tmp*tmp;

  // add ||r_pri||^2
  tmp = x.sum();
  rt += tmp*tmp;

  // add ||r_cent||^2
  for (uint i=0; i<n-1; ++i) {
    tmp = -lambda(i)*x(i) - 1/t;
    rt += tmp*tmp;
  }
  tmp = lamp*x(n-1) - 1/t;
  rt += tmp*tmp;
  tmp = lamn*(C-x(n-1)) - 1/t;
  rt += tmp*tmp;
  
  return(sqrt(rt));
}


// this function checks if all constarints are satisfied by x
bool satisfy_inequality_constraints(infra::vector_base x,double C) {

  for (uint i=0; i<x.size()-1; ++i) {
    if (x(i) >= 0.0) return(false);
  }
  if (x(x.size()-1) <= 0.0 || x(x.size()-1) >= C) return(false);

  return(true);
}

