//=============================================================================
// File Name: mcSparse.cc
// Written by: Shai Shalev-Shwartz (23.03.06)
// This code implements three multiclass online learning algorithm:
//   - Perceptron with max-update
//   - PA-I with max-update
//   - SOPOPO for multiclass
// The implementation is primal with sparse vectors
//=============================================================================
#ifndef _SHAI_MCSPARSE_
#define _SHAI_MCSPARSE_

//*****************************************************************************
// Included Files
//*****************************************************************************
#include <iostream>
#include <fstream>
#include <stdio.h>
#include <cstdlib>
#include <vector>
#include <cmdline/cmd_line.h>
#include <infra.h>
#include <simple_sparse_vector/simple_sparse_vector.h>
#include <simple_sparse_vector/simple_sparse_vector.imp>

#define _min(x,y) ( ((x) < (y)) ? (x) : (y) )


void EG_IPsolver(infra::vector_base P, infra::vector_base Q, 
		 double gamma, double C, unsigned int y,
		 infra::vector_base x) ;


int main(int argc, char** argv) {

  // PARSE COMMAND LINE
  // ====================================================================
  // check for the correct number of command line arguments and print usage
  // message
  std::string data_filename;
  std::string label_filename;
  std::string method;
  double C,gamma;
  std::string out_file;
  std::string log_file_name;
  uint max_examples;
  uint num_labels;
  uint n;
  bool make_binary = true;

  // parse command line
  learning::cmd_line cmdline;
  cmdline.info("Online multiclass learning - EG primal sparse binary");
  cmdline.add_master_option("<instance file name (infra format)>", 
			    &data_filename);
  cmdline.add_master_option("<labels file name (ascii numbers from 0,...,k-1)>", 
			    &label_filename);
  cmdline.add("-C", "value of the C parameter [default 1.0]", &C, 1.0);
  cmdline.add("-gamma", "value of the gamma parameter [default 0.05]", &gamma, 0.05);
  cmdline.add("-n", "the dimension of examples [default 10000]", &n, 10000);
  cmdline.add("-m","maximal number of examples to read from the file\n "\
	      "\t[default ten milion (10000000)]",
	      &max_examples,
	      10000000);
  cmdline.add("-log","Name of a file in which on each line there is the number of trial,\n "\
	      "\tthe prediction, the number of cumulative mistakes and cumulative loss.\n"\
	      "\t[default no_log]",
	      &log_file_name,
	      "no_log");
  cmdline.add("-method", "The online method to use: \n\tPerc for Perceptron, "\
	      "PA for Passive-Aggressive, "\
	      "Sop for SOPOPO\n [Default: PA]",
	      &method,
	      "PA");
//   cmdline.add("-b","Use this flag for converting instances to binary instances\n",
// 	      &make_binary,false);

  int rc = cmdline.parse(argc, argv);
  if (rc < 2) {
    cmdline.print_help();
    return EXIT_FAILURE;
  }
  if (method != "Perc" && method != "PA" && method != "Sop") {
    std::cerr << "incorrect method type. Abort " << std::endl;
    return EXIT_FAILURE;
  }

  

  // OPEN DATA AND LABEL FILES
  // =========================
  std::ifstream data_file(data_filename.c_str());
  if (!data_file.good()) {
    std::cerr << "error w/ " << data_filename << std::endl;
    return EXIT_FAILURE;
  }
  std::ifstream label_file(label_filename.c_str());
  if (!label_file.good()) {
    std::cerr << "error w/ label_file" << std::endl;
    return EXIT_FAILURE;
  }


  // open log file
  FILE * log_fid = NULL;
  if (log_file_name != "no_log")
    log_fid = fopen(log_file_name.c_str(),"wt");


  // INITIALIZATIONS
  // ===============
  // read the number of lables. This is the first number in the labels file
  label_file >> num_labels;

  // allocate an stl vector of the exp(theta) vectors for each label, where if
  // theta_j = 0 we don't save it in exptheta
  std::vector< simple_sparse_vector > exptheta(num_labels); 
  // allocate an infra vector of sum_j exp(theta_j) for each label
  infra::vector sumexptheta(num_labels); 
  sumexptheta = n; // initially, theta_j = 0 and therefore sumexptheta is n

  // Count mistakes, and loss
  unsigned long mistakes = 0;
  double loss = 0.0;


  unsigned long svc = 0;

  // BEGIN ONLINE LEARNING
  // =====================
  unsigned long rounds = 0;
  while (rounds < max_examples && data_file.good() && label_file.good()) {

    // Receive a new instance from the data file
    // =========================================
    simple_sparse_vector instance(data_file);
    if (make_binary) instance.make_binary();

    // predict a label 
    // ===============
    infra::vector ranks(num_labels); 
    ranks.zeros();
    for(std::map<std::string, double>::iterator it = instance.my_vec.begin(); 
	it != instance.my_vec.end(); it++) {
      for (unsigned int i=0; i < num_labels; ++i) {
	// check if exptheta[i] contains the key
	if (exptheta[i].my_vec.count((*it).first) != 0) 
	  ranks(i) += (*it).second * exptheta[i].my_vec[(*it).first] / sumexptheta(i);
	else
	  ranks(i) += (*it).second / sumexptheta(i);
      }
    }
    uint predicted_label = ranks.argmax();


    // Read the correct label from the labels file
    // ===========================================
    uint correct_label;
    label_file >> correct_label;

    // calculate the loss
    // ===================

    // first, find the highest incorrect label
    uint highest_ranked_incorrect = predicted_label;
    if (correct_label == predicted_label) {
      double tmp = ranks(correct_label);
      ranks(correct_label) = ranks.min() - 1000000.0;
      highest_ranked_incorrect = ranks.argmax();
      ranks(correct_label) = tmp;
    }
    else {
      mistakes++;
    }
    
    // calculate the margin of the classification
    double margin = ranks(correct_label) - ranks(highest_ranked_incorrect);

    // and the loss is
    double cur_loss = gamma - margin; 
    if (cur_loss < 0.0) cur_loss = 0.0;

    loss += cur_loss; // the cumulative loss

    // update
    // ==========================

    // For Perceptron, update only if there's a mistake (cur_loss >= gamma)
    // For PA and Sop, update whenever cur_loss > 0.0
    if ( ((method != "Perc" && cur_loss > 0.0) || (cur_loss >= gamma)) && instance.my_vec.size() != 0) { 
      
      // Perceptron and PA update
      if (method == "Perc" || method == "PA") {

	// find the alpha
	double alpha = C; // Perceptron
	if (method == "PA") { // PA-I


#ifdef  UNEXPLAINED_EG

	  // we normalize by a bound on <x,H x> which is sum_j w_j x_j^2
	  alpha = cur_loss / (ranks(correct_label) + ranks(highest_ranked_incorrect));

#else  // Correct EG

	  double sum_w_correct = 0.0;
	  double sum_w_incorrect = 0.0;
	  for(std::map<std::string, double>::iterator it = instance.my_vec.begin(); 
	      it != instance.my_vec.end(); it++) {

	    if (exptheta[correct_label].my_vec.count((*it).first) != 0) 
	      sum_w_correct += exptheta[correct_label].my_vec[(*it).first];
	    else
	      sum_w_correct += 1.0;

	    if (exptheta[highest_ranked_incorrect].my_vec.count((*it).first) != 0) 
	      sum_w_incorrect += exptheta[highest_ranked_incorrect].my_vec[(*it).first];
	    else
	      sum_w_incorrect += 1.0;

	  }

	  double a = (sumexptheta(correct_label)-sum_w_correct)/sum_w_correct;
	  double b = (sumexptheta(highest_ranked_incorrect)-sum_w_incorrect)/
	    sum_w_incorrect;
	  double A = b*(gamma-1);
	  double B = gamma*(1+a*b);
	  double CC = (1+gamma)*a;
	  double sqrtDelta = sqrt(B*B-4*A*CC);
	  double beta = (-B-sqrtDelta)/(2*A);
	  alpha = log(beta); 

#endif

	  if (alpha > C) {
	    alpha = C;
	    svc++;
	  }

	  if (alpha < 0.0) {
	    std::cerr << "alpha < 0.0. Probably a bug here. Aborting" << std::endl;
	    return(EXIT_FAILURE);
	  }
	}  // end if (method == "PA")


	// after we found alpha (either Perceptron or PA)
	// we update exptheta and sumexptheta
	for(std::map<std::string, double>::iterator it = instance.my_vec.begin(); 
	    it != instance.my_vec.end(); it++) {

	  double tmp = exp(alpha * (*it).second);

	  // This is like theta[correct_label].add(instance,alpha) 
	  
	  if (exptheta[correct_label].my_vec.count((*it).first) != 0) {
	    sumexptheta(correct_label) -= exptheta[correct_label].my_vec[(*it).first];
	    exptheta[correct_label].my_vec[(*it).first] *= tmp;
	    sumexptheta(correct_label) += exptheta[correct_label].my_vec[(*it).first];
	  } else {
	    sumexptheta(correct_label) += (tmp - 1);
	    exptheta[correct_label].my_vec[(*it).first] = tmp;
	  }

	  // This is like theta[highest_ranked_incorrect].add(instance,-alpha) 
	  tmp = exp(-alpha * (*it).second);
	  if (exptheta[highest_ranked_incorrect].my_vec.count((*it).first) != 0) {
	    sumexptheta(highest_ranked_incorrect) -= exptheta[highest_ranked_incorrect].my_vec[(*it).first];
	    exptheta[highest_ranked_incorrect].my_vec[(*it).first] *= tmp;
	    sumexptheta(highest_ranked_incorrect) += exptheta[highest_ranked_incorrect].my_vec[(*it).first];
	  } else {
	    sumexptheta(highest_ranked_incorrect) += (tmp - 1);
	    exptheta[highest_ranked_incorrect].my_vec[(*it).first] = tmp;
	  }
	}
	
	
      } else { 	// Sopopo update
	
	// allocate alpha and initialize it with the Perceptron update
	infra::vector alpha(num_labels); 
	alpha.zeros(); 
	alpha(correct_label) = C; 
	alpha(highest_ranked_incorrect) = -C;
	
	
	// for efficiency, we prepare the following two vectors of size num_labels:
	// 1) mu(r) = ranks(r)*sumexptheta(r) = sum_{i : x_i =1} exp(theta_i)
	// 2) nu(r) = sumexptheta(r) - mu(r)
	infra::vector mu(ranks); mu *= sumexptheta;
	infra::vector nu(sumexptheta); nu -= mu;


	//std::cerr << "alpha before:\n" << alpha << std::endl;	
	
	// call the EG IP solver
	EG_IPsolver(mu,nu,gamma,C,correct_label,alpha);
	
	//std::cerr << "alpha after:\n" << alpha << std::endl;	
	
	// Finally, update exptheta and sumexptheta
	for(std::map<std::string, double>::iterator it = instance.my_vec.begin(); 
	    it != instance.my_vec.end(); it++) {
	  for (unsigned int i=0; i < num_labels; ++i) {
	    if (fabs(alpha(i)) > 1e-6) {
	      // This is like theta[i].add(instance,alpha(i)) 
	      double tmp = exp(alpha(i));
	      if (exptheta[i].my_vec.count((*it).first) != 0) {
		sumexptheta(i) -= exptheta[i].my_vec[(*it).first];
		exptheta[i].my_vec[(*it).first] *= tmp;
		sumexptheta(i) += exptheta[i].my_vec[(*it).first];
	      } else {
		sumexptheta(i) += (tmp - 1);
		exptheta[i].my_vec[(*it).first] = tmp;
	      }
	    }
	  }
	}
	
	
      }
    }
    
    // print some statistics
    if (log_fid != NULL)
      fprintf(log_fid,"rounds: %lu mistakes: %lu loss %.3f svc %lu\n",
	      rounds,mistakes,loss,svc);
    
    // and increment the rounds
    rounds++;
    
  }
  
  // PRINT OUT THE FINAL REPORT (do not change without discussing)
  // ============================================================
  std::cout << "method: " << method
	    << "\tC: " << C
	    << "\trounds: " << rounds
	    << "\tmistakes: " << mistakes
	    << "\tloss: " << loss
	    << "\terror_rate: " << double(mistakes)/double(rounds)
    	    << "\tavg_loss: " << loss/double(rounds) 
            << "\trep_set_size: " << svc
	    << std::endl; 


  // close log file
  if (log_fid != NULL) 
    fclose(log_fid);

  return(EXIT_SUCCESS);

}


#endif
//*****************************************************************************
//                                     E O F
//*****************************************************************************

